from acoustics.directivity import *
import numpy as np
import pytest


@pytest.mark.parametrize("given, expected, uncertainty", [
    (0.0, 1.0, 0.0),
    (1. / 2. * np.pi, 0.0, 0.0),
    (np.pi, +1.0, 0.0),
    (3. / 2. * np.pi, 0.0, 0.0),
    (2. * np.pi, +1.0, 0.0),
])
def test_figure_eight(given, expected, uncertainty):
    assert figure_eight(given) == pytest.approx(expected, uncertainty)



"""
Tests for :class:`Acoustics.Octave.Octave`
"""

from acoustics.octave import Octave
import numpy as np


class TestOctave():
    def test_interval(self):
        emin = 1.0
        emax = 4.0
        f = np.logspace(emin, emax, 50)

        o = Octave(interval=f)

        assert (o.fmin == 10.0**emin)
        assert (o.fmax == 10.0**emax)
        assert (len(o.n) == len(o.center))

        o.unique = True
        assert (len(o.n) == len(f))

    def test_minmax(self):
        fmin = 250.0
        fmax = 1000.0

        o = Octave(fmin=fmin, fmax=fmax)

        assert (len(o.center) == 3)  # 250, 500, and 1000 Hz
        assert (len(o.n) == 3)



import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal

import pytest

from acoustics.room import (mean_alpha, nrc, t60_sabine, t60_eyring, t60_millington, t60_fitzroy, t60_arau, t60_impulse,
                            c50_from_file, c80_from_file)
from acoustics.bands import octave, third

import sys
sys.path.append('..')
from get_data_path import data_path


def setup_module(room):
    room.surfaces = np.array([240, 600, 500])
    room.alpha = np.array([0.1, 0.25, 0.45])
    room.alpha_bands = np.array([[0.1, 0.1, 0.1, 0.1], [0.25, 0.25, 0.25, 0.25], [0.45, 0.45, 0.45, 0.45]])
    room.volume = 3000


def test_t60_sabine():
    calculated = t60_sabine(surfaces, alpha, volume)
    real = 1.211382149
    assert_almost_equal(calculated, real)


def test_t60_sabine_bands():
    calculated = t60_sabine(surfaces, alpha_bands, volume)
    real = np.array([1.211382149, 1.211382149, 1.211382149, 1.211382149])
    assert_array_almost_equal(calculated, real)


def test_t60_eyring():
    calculated = t60_eyring(surfaces, alpha, volume)
    real = 1.020427763
    assert_almost_equal(calculated, real)


def test_t60_eyring_bands():
    calculated = t60_eyring(surfaces, alpha_bands, volume)
    real = np.array([1.020427763, 1.020427763, 1.020427763, 1.020427763])
    assert_array_almost_equal(calculated, real)


def test_t60_millington():
    calculated = t60_millington(surfaces, alpha, volume)
    real = 1.020427763
    assert_almost_equal(calculated, real)


def test_t60_millington_bands():
    calculated = t60_millington(surfaces, alpha_bands, volume)
    real = np.array([1.020427763, 1.020427763, 1.020427763, 1.020427763])
    assert_array_almost_equal(calculated, real)


def test_t60_fitzroy():
    surfaces_fitzroy = np.array([240, 240, 600, 600, 500, 500])
    alpha_fitzroy = np.array([0.1, 0.1, 0.25, 0.25, 0.45, 0.45])
    calculated = t60_fitzroy(surfaces_fitzroy, alpha_fitzroy, volume)
    real = 0.699854185
    assert_almost_equal(calculated, real)


def test_t60_fitzroy_bands():
    surfaces_fitzroy = np.array([240, 240, 600, 600, 500, 500])
    alpha_bands_f = np.array([[0.1, 0.1, 0.25, 0.25, 0.45, 0.45], [0.1, 0.1, 0.25, 0.25, 0.45, 0.45],
                              [0.1, 0.1, 0.25, 0.25, 0.45, 0.45]])
    calculated = t60_fitzroy(surfaces_fitzroy, alpha_bands_f, volume)
    real = np.array([0.699854185, 0.699854185, 0.699854185])
    assert_array_almost_equal(calculated, real)


def test_t60_arau():
    Sx = surfaces[0]
    Sy = surfaces[1]
    Sz = surfaces[2]
    calculated = t60_arau(Sx, Sy, Sz, alpha, volume)
    real = 1.142442931
    assert_almost_equal(calculated, real)


def test_t60_arau_bands():
    Sx = surfaces[0]
    Sy = surfaces[1]
    Sz = surfaces[2]
    calculated = t60_arau(Sx, Sy, Sz, alpha_bands, volume)
    real = np.array([1.142442931, 1.142442931, 1.142442931, 1.142442931])
    assert_array_almost_equal(calculated, real)


def test_mean_alpha_float():
    alpha = 0.1
    surface = 10
    calculated = mean_alpha(alpha, surface)
    real = 0.1
    assert_almost_equal(calculated, real)


def test_mean_alpha_1d():
    alpha = np.array([0.1, 0.2, 0.3])
    surfaces = np.array([20, 30, 40])
    calculated = mean_alpha(alpha, surfaces)
    real = 0.222222222
    assert_almost_equal(calculated, real)


def test_mean_alpha_bands():
    alpha = np.array([[0.1, 0.1, 0.1], [0.2, 0.2, 0.2], [0.3, 0.3, 0.3]])
    surfaces = np.array([20, 30, 40])
    calculated = mean_alpha(alpha, surfaces)
    real = np.array([0.222222222, 0.222222222, 0.222222222])
    assert_array_almost_equal(calculated, real)


def test_nrc_1d():
    alpha = np.array([0.1, 0.25, 0.5, 0.9])
    calculated = nrc(alpha)
    real = 0.4375
    assert_almost_equal(calculated, real)


def test_nrc_2d():
    alphas = np.array([[0.1, 0.2, 0.3, 0.4], [0.4, 0.5, 0.6, 0.7]])
    calculated = nrc(alphas)
    real = np.array([0.25, 0.55])
    assert_array_almost_equal(calculated, real)


@pytest.mark.parametrize("file_name, bands, rt, expected", [
    (data_path() + 'ir_sportscentre_omni.wav', octave(125, 4000), 't30',
     np.array([7.388, 8.472, 6.795, 6.518, 4.797, 4.089])),
    (data_path() + 'ir_sportscentre_omni.wav', octave(125, 4000), 'edt',
     np.array([4.667, 5.942, 6.007, 5.941, 5.038, 3.735])),
    (data_path() + 'living_room_1.wav', octave(63, 8000), 't30',
     np.array([0.274, 0.365, 0.303, 0.259, 0.227, 0.211, 0.204, 0.181])),
    (data_path() + 'living_room_1.wav', octave(63, 8000), 't20',
     np.array([0.300, 0.365, 0.151, 0.156, 0.102, 0.076, 0.146, 0.152])),
    (data_path() + 'living_room_1.wav', octave(63, 8000), 't10',
     np.array([0.185, 0.061, 0.109, 0.024, 0.039, 0.023, 0.105, 0.071])),
    (data_path() + 'living_room_1.wav', octave(63, 8000), 'edt',
     np.array([0.267, 0.159, 0.080, 0.037, 0.021, 0.010, 0.022, 0.020])),
    (data_path() + 'living_room_1.wav', third(100, 5000), 't30',
     np.array([
         0.318, 0.340, 0.259, 0.311, 0.267, 0.376, 0.342, 0.268, 0.212, 0.246, 0.211, 0.232, 0.192, 0.231, 0.252, 0.202,
         0.184, 0.216
     ])),
    (data_path() + 'living_room_1.wav', third(100, 5000), 't20',
     np.array([
         0.202, 0.383, 0.189, 0.173, 0.141, 0.208, 0.323, 0.221, 0.102, 0.110, 0.081, 0.128, 0.072, 0.074, 0.087, 0.129,
         0.137, 0.171
     ])),
    (data_path() + 'living_room_1.wav', third(100, 5000), 't10',
     np.array([
         0.110, 0.104, 0.132, 0.166, 0.135, 0.040, 0.119, 0.223, 0.025, 0.023, 0.047, 0.050, 0.010, 0.017, 0.039, 0.084,
         0.154, 0.093
     ])),
    (data_path() + 'living_room_1.wav', third(100, 5000), 'edt',
     np.array([
         0.354, 0.328, 0.284, 0.210, 0.132, 0.116, 0.085, 0.114, 0.064, 0.045, 0.047, 0.047, 0.024, 0.017, 0.016, 0.022,
         0.020, 0.036
     ])),
])
def test_t60_impulse(file_name, bands, rt, expected):
    calculated = t60_impulse(file_name, bands, rt)
    assert_array_almost_equal(calculated, expected, decimal=0)


@pytest.mark.parametrize("file_name, bands, expected", [
    (data_path() + 'living_room_1.wav', octave(63, 8000), np.array([8., 18., 23., 26., 30., 31., 27., 29.])),
    (data_path() + 'living_room_1.wav', third(100, 5000),
     np.array([3., 6., 7., 13., 18., 23., 20., 19., 28., 30., 30., 27., 32., 31., 30., 28., 29., 25.])),
])
def test_c50_from_file(file_name, bands, expected):
    calculated = c50_from_file(file_name, bands)
    assert_array_almost_equal(calculated, expected, decimal=0)


@pytest.mark.parametrize("file_name, bands, expected", [
    (data_path() + 'living_room_1.wav', octave(63, 8000),
     np.array([18.542, 23.077, 27.015, 31.743, 35.469, 36.836, 33.463, 36.062])),
    (data_path() + 'living_room_1.wav', third(100, 5000),
     np.array([17., 14., 17., 24., 26., 27., 22., 26., 34., 35., 34., 32., 38., 38., 34., 34., 35., 32.])),
])
def test_c80_from_file(file_name, bands, expected):
    calculated = c80_from_file(file_name, bands)
    assert_array_almost_equal(calculated, expected, decimal=0)


def teardown_module(room):
    pass



from acoustics.decibel import *


def test_dbsum():
    assert (abs(dbsum([10.0, 10.0]) - 13.0103) < 1e-5)


def test_dbmean():
    assert (dbmean([10.0, 10.0]) == 10.0)


def test_dbadd():
    assert (abs(dbadd(10.0, 10.0) - 13.0103) < 1e-5)


def test_dbsub():
    assert (abs(dbsub(13.0103, 10.0) - 10.0) < 1e-5)


def test_dbmul():
    assert (abs(dbmul(10.0, 2) - 13.0103) < 1e-5)


def test_dbdiv():
    assert (abs(dbdiv(13.0103, 2) - 10.0) < 1e-5)



import pytest
import numpy as np
from acoustics.generator import noise
from acoustics.signal import octaves


class Test_noise():
    """
    Test :func:`acoustics.generator.noise`.
    """
    parameters = [
        ('white', +3.0, 0.0),
        ('pink', 0.0, -3.0),
        ('blue', +6.0, +3.0),
        ('brown', -3.0, -6.0),
        ('violet', +9.0, +6.0),
    ]
    # color, power_change, power_density_change

    ERROR = 1.0
    """Permitted error for power_change and power_density_change.
    This value is relatively large for `violet` because of its steep
    inclination. On other cases the error is mostly < 0.1
    """

    @pytest.fixture(params=parameters)
    def parameters(self, request):
        return request.param

    @pytest.fixture
    def color(self, parameters):
        return parameters[0]

    @pytest.fixture
    def power_change(self, parameters):
        return parameters[1]

    @pytest.fixture
    def power_density_change(self, parameters):
        return parameters[2]

    @pytest.fixture(params=[48000 * 10, 48000 * 10 + 1])
    def samples(self, request):
        return request.param

    def test_length(self, color, samples):

        assert (len(noise(samples, color)) == samples)

    def test_power(self, color, samples, power_change):

        fs = 48000
        _, L = octaves(noise(samples, color), fs)
        change = np.diff(L).mean()
        assert (np.abs(change - power_change) < self.ERROR)

    def test_power_density(self, color, samples, power_density_change):

        fs = 48000
        _, L = octaves(noise(samples, color), fs, density=True)
        change = np.diff(L).mean()
        assert (np.abs(change - power_density_change) < self.ERROR)



"""
Ambisonics
==========

"""

import numpy as np
try:
    from scipy import math
    factorial = math.factorial
except AttributeError:
    from scipy.misc import factorial


def acn(order=1):
    """Spherical harmonic degree `n` and order `m` for ambisonics order `order`.

    :param order: Ambisonics order.

    Yields tuples `(n, m)` where `n` is the degree and `m` the order.

    Follows ACN.

    === == == ======
    ACN n  m  letter
    === == == ======
    0   0  0  W
    --- -- -- ------
    1   1  -1 Y
    --- -- -- ------
    2   1  0  Z
    --- -- -- ------
    3   1  +1 X
    === == == ======

    """
    for n in range(order + 1):
        for m in range(-n, n + 1):
            yield (n, m)


def sn3d(m, n):
    """SN3D or Schmidt semi-normalisation

    :param m: order `n`
    :param n: degree `m`

    http://en.wikipedia.org/wiki/Ambisonic_data_exchange_formats#SN3D

    """
    m = np.atleast_1d(m)
    n = np.atleast_1d(n)

    d = np.logical_not(m.astype(bool))
    out = np.sqrt((2.0 - d) / (4.0 * np.pi) * factorial(n - np.abs(m)) / factorial(n + np.abs(m)))
    return out


def n3d(m, n):
    """N3D or full three-D normalisation

    :param m: order `n`
    :param n: degree `m`

    http://en.wikipedia.org/wiki/Ambisonic_data_exchange_formats#N3D

    """
    n = np.atleast_1d(n)
    return sn3d(m, n) * np.sqrt(2 * n + 1)


__all__ = ['acn', 'sn3d', 'n3d']



import numpy as np

import pytest

has_matplotlib = pytest.importorskip("matplotlib")

if has_matplotlib:
    from acoustics.bands import octave, third
    from acoustics.imaging import plot_octave, plot_third, plot_bands


def setup_module(imaging):
    imaging.octaves = octave(16, 16000)
    imaging.thirds = third(63, 8000)
    imaging.tl_oct = np.array([3, 4, 5, 12, 15, 24, 28, 23, 35, 45, 55])
    imaging.tl_third = np.array([0, 0, 0, 1, 1, 2, 3, 5, 8, 13, 21, 32, 41, 47, 46, 44, 58, 77, 61, 75, 56, 54])
    imaging.title = 'Title'
    imaging.label = 'Label'


def test_plot_octave():
    plot_octave(tl_oct, octaves)


def test_plot_octave_kHz():
    plot_octave(tl_oct, octaves, kHz=True, xlabel=label, ylabel=label, title=title, separator='.')


def test_plot_third_octave():
    plot_third(tl_third, thirds, marker='s', separator=',')


def test_plot_third_octave_kHz():
    plot_third(tl_third, thirds, marker='s', kHz=True, xlabel=label, ylabel=label, title=title)


def test_plot_band_oct():
    plot_bands(tl_oct, octaves, axes=None, band_type='octave')


def teardown_module(imaging):
    pass



"""
Descriptors
============

The descriptors module offers all kinds of acoustics related descriptors.

.. toctree::
    :maxdepth: 2

Descriptors from ISO/TR 25417:2007
**********************************

Descriptors from :mod:`acoustics.standards.iso_tr_25417_2007`.

.. autoattribute:: acoustics.descriptors.REFERENCE_PRESSURE
.. autofunction:: acoustics.descriptors.sound_pressure_level
.. autofunction:: acoustics.descriptors.equivalent_sound_pressure_level
.. autofunction:: acoustics.descriptors.peak_sound_pressure
.. autofunction:: acoustics.descriptors.peak_sound_pressure_level
.. autoattribute:: acoustics.descriptors.REFERENCE_SOUND_EXPOSURE
.. autofunction:: acoustics.descriptors.sound_exposure
.. autofunction:: acoustics.descriptors.sound_exposure_level
.. autoattribute:: acoustics.descriptors.REFERENCE_POWER
.. autofunction:: acoustics.descriptors.sound_power_level
.. autofunction:: acoustics.descriptors.sound_energy
.. autofunction:: acoustics.descriptors.sound_energy_level
.. autoattribute:: acoustics.descriptors.REFERENCE_ENERGY
.. autofunction:: acoustics.descriptors.sound_intensity
.. autofunction:: acoustics.descriptors.time_averaged_sound_intensity
.. autoattribute:: acoustics.descriptors.REFERENCE_INTENSITY
.. autofunction:: acoustics.descriptors.time_averaged_sound_intensity_level
.. autofunction:: acoustics.descriptors.normal_time_averaged_sound_intensity
.. autofunction:: acoustics.descriptors.normal_time_averaged_sound_intensity_level


Other descriptors
*****************

"""
import numpy as np

from acoustics.standards.iso_tr_25417_2007 import (
    REFERENCE_PRESSURE,
    sound_pressure_level,
    equivalent_sound_pressure_level,
    peak_sound_pressure,
    peak_sound_pressure_level,
    REFERENCE_SOUND_EXPOSURE,
    sound_exposure,
    sound_exposure_level,
    REFERENCE_POWER,
    sound_power_level,
    sound_energy,
    REFERENCE_ENERGY,
    sound_energy_level,
    sound_intensity,
    time_averaged_sound_intensity,
    REFERENCE_INTENSITY,
    time_averaged_sound_intensity_level,
    normal_time_averaged_sound_intensity,
    normal_time_averaged_sound_intensity_level,
)

from acoustics.standards.iso_1996_1_2003 import composite_rating_level


def _leq(levels, time):
    levels = np.asarray(levels)
    return 10.0 * np.log10((1.0 / time) * np.sum(10.0**(levels / 10.0)))


def leq(levels, int_time=1.0):
    """
    Equivalent level :math:`L_{eq}`.

    :param levels: Levels as function of time.
    :param int_time: Integration time. Default value is 1.0 second.
    :returns: Equivalent level L_{eq}.

    Sum of levels in dB.
    """
    levels = np.asarray(levels)
    time = levels.size * int_time
    return _leq(levels, time)


def sel(levels):
    """
    Sound Exposure Level from ``levels`` (NumPy array).
    """
    levels = np.asarray(levels)
    return _leq(levels, 1.0)


def lw(W, Wref=1.0e-12):
    """
    Sound power level :math:`L_{w}` for sound power :math:`W` and reference power :math:`W_{ref}`.

    :param W: Sound power :math:`W`.
    :param Wref: Reference power :math:`W_{ref}`. Default value is :math:`10^{12}` watt.
    """
    W = np.asarray(W)
    return 10.0 * np.log10(W / Wref)


def lden(lday, levening, lnight, hours=(12.0, 4.0, 8.0), adjustment=(0.0, 5.0, 10.0)):
    """
    Calculate :math:`L_{den}` from :math:`L_{day}`, :math:`L_{evening}` and :math:`L_{night}`.

    :param lday: Equivalent level during day period :math:`L_{day}`.
    :param levening: Equivalent level during evening period :math:`L_{evening}`.
    :param lnight: Equivalent level during night period :math:`L_{night}`.
    :param hours: Hours per period.
    :param adjustment: Correction factor per period.
    :returns: :math:`L_{den}`

    .. seealso:: :func:`acoustics.standards.iso_1996_1_2003.composite_rating_level`
    """
    lday = np.asarray(lday)
    levening = np.asarray(levening)
    lnight = np.asarray(lnight)
    return composite_rating_level(np.vstack((lday, levening, lnight)).T, hours, adjustment)


def ldn(lday, lnight, hours=(15.0, 9.0), adjustment=(0.0, 10.0)):
    """
    Calculate :math:`L_{dn}` from :math:`L_{day}` and :math:`L_{night}`.

    :param lday: Equivalent level during day period :math:`L_{day}`.
    :param lnight: Equivalent level during night period :math:`L_{night}`.
    :param hours: Hours per period.
    :param adjustment: Correction factor per period.
    :returns: :math:`L_{dn}`

    .. seealso:: :func:`acoustics.standards.iso_1996_1_2003.composite_rating_level`
    """
    lday = np.asarray(lday)
    lnight = np.asarray(lnight)
    return composite_rating_level(np.vstack((lday, lnight)).T, hours, adjustment)


__all__ = [
    # Following we take from another module
    "REFERENCE_PRESSURE",
    "sound_pressure_level",
    "equivalent_sound_pressure_level",
    "peak_sound_pressure",
    "peak_sound_pressure_level",
    "REFERENCE_SOUND_EXPOSURE",
    "sound_exposure",
    "sound_exposure_level",
    "REFERENCE_POWER",
    "sound_power_level",
    "sound_energy",
    "REFERENCE_ENERGY",
    "sound_energy_level",
    "sound_intensity",
    "time_averaged_sound_intensity",
    "REFERENCE_INTENSITY",
    "time_averaged_sound_intensity_level",
    "normal_time_averaged_sound_intensity",
    "normal_time_averaged_sound_intensity_level",
    # Following are locally defined
    "leq",
    "sel",
    "lw",
    "lden",
    "ldn",
]



"""
Octave
======

Module for working with octaves.

The following is an example on how to use :class:`acoustics.octave.Octave`.

.. literalinclude:: ../examples/octave.py

"""
import numpy as np
import acoustics

#REFERENCE = 1000.0
#"""
#Reference frequency.
#"""

from acoustics.standards.iec_61260_1_2014 import index_of_frequency
from acoustics.standards.iec_61260_1_2014 import REFERENCE_FREQUENCY as REFERENCE


def exact_center_frequency(frequency=None, fraction=1, n=None, ref=REFERENCE):
    """Exact center frequency.

    :param frequency: Frequency within the band.
    :param fraction: Band designator.
    :param n: Index of band.
    :param ref: Reference frequency.
    :return: Exact center frequency for the given frequency or band index.

    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.exact_center_frequency`
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.index_of_frequency`

    """
    if frequency is not None:
        n = acoustics.standards.iec_61260_1_2014.index_of_frequency(frequency, fraction=fraction, ref=ref)
    return acoustics.standards.iec_61260_1_2014.exact_center_frequency(n, fraction=fraction, ref=ref)


def nominal_center_frequency(frequency=None, fraction=1, n=None):
    """Nominal center frequency.

    :param frequency: Frequency within the band.
    :param fraction: Band designator.
    :param n: Index of band.
    :returns: The nominal center frequency for the given frequency or band index.

    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.exact_center_frequency`
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.nominal_center_frequency`

    .. note:: Contrary to the other functions this function silently assumes 1000 Hz reference frequency.

    """
    center = exact_center_frequency(frequency, fraction, n)
    return acoustics.standards.iec_61260_1_2014.nominal_center_frequency(center, fraction)


def lower_frequency(frequency=None, fraction=1, n=None, ref=REFERENCE):
    """Lower band-edge frequency.

    :param frequency: Frequency within the band.
    :param fraction: Band designator.
    :param n: Index of band.
    :param ref: Reference frequency.
    :returns: Lower band-edge frequency for the given frequency or band index.

    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.exact_center_frequency`
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.lower_frequency`

    """
    center = exact_center_frequency(frequency, fraction, n, ref=ref)
    return acoustics.standards.iec_61260_1_2014.lower_frequency(center, fraction)


def upper_frequency(frequency=None, fraction=1, n=None, ref=REFERENCE):
    """Upper band-edge frequency.

    :param frequency: Frequency within the band.
    :param fraction: Band designator.
    :param n: Index of band.
    :param ref: Reference frequency.
    :returns: Upper band-edge frequency for the given frequency or band index.

    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.exact_center_frequency`
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.upper_frequency`

    """
    center = exact_center_frequency(frequency, fraction, n, ref=ref)
    return acoustics.standards.iec_61260_1_2014.upper_frequency(center, fraction)


#-- things below will be deprecated?---#

frequency_of_band = acoustics.standards.iec_61260_1_2014.exact_center_frequency
band_of_frequency = index_of_frequency


class Octave:
    """
    Class to calculate octave center frequencies.
    """

    def __init__(self, fraction=1, interval=None, fmin=None, fmax=None, unique=False, reference=REFERENCE):

        self.reference = reference
        """
        Reference center frequency :math:`f_{c,0}`.
        """

        self.fraction = fraction
        """
        Fraction of octave.
        """

        if (interval is not None) and (fmin is not None or fmax is not None):
            raise AttributeError

        self._interval = np.asarray(interval)
        """Interval"""

        self._fmin = fmin
        """Minimum frequency of a range."""

        self._fmax = fmax
        """Maximum frequency of a range."""

        self.unique = unique
        """Whether or not to calculate the requested values for every value of ``interval``."""

    @property
    def fmin(self):
        """Minimum frequency of an interval."""
        if self._fmin is not None:
            return self._fmin
        elif self._interval is not None:
            return self.interval.min()
        else:
            raise ValueError("Incorrect fmin/interval")

    @fmin.setter
    def fmin(self, x):
        if self.interval is not None:
            pass  # Warning, remove interval first.
        else:
            self._fmin = x

    @property
    def fmax(self):
        """Maximum frequency of an interval."""
        if self._fmax is not None:
            return self._fmax
        elif self._interval is not None:
            return self.interval.max()
        else:
            raise ValueError("Incorrect fmax/interval")

    @fmax.setter
    def fmax(self, x):
        if self.interval is not None:
            pass
        else:
            self._fmax = x

    @property
    def interval(self):
        """Interval."""
        return self._interval

    @interval.setter
    def interval(self, x):
        if self._fmin or self._fmax:
            pass
        else:
            self._interval = np.asarray(x)

    def _n(self, f):
        """
        Calculate the band ``n`` from a given frequency.

        :param f: Frequency

        See also :func:`band_of_frequency`.
        """
        return band_of_frequency(f, fraction=self.fraction, ref=self.reference)

    def _fc(self, n):
        """
        Calculate center frequency of band ``n``.

        :param n: band ``n`.

        See also :func:`frequency_of_band`.
        """
        return frequency_of_band(n, fraction=self.fraction, ref=self.reference)

    @property
    def n(self):
        """
        Return band ``n`` for a given frequency.
        """
        if self.interval is not None and self.unique:
            return self._n(self.interval)
        else:
            return np.arange(self._n(self.fmin), self._n(self.fmax) + 1)

    @property
    def center(self):
        r"""
        Return center frequencies :math:`f_c`.

        .. math::  f_c = f_{ref} \cdot 2^{n/N} \cdot 10^{\frac{3}{10N}}

        """
        n = self.n
        return self._fc(n)

    @property
    def bandwidth(self):
        """
        Bandwidth of bands.

        .. math:: B = f_u - f_l

        """
        return self.upper - self.lower

    @property
    def lower(self):
        r"""
        Lower frequency limits of bands.

        .. math:: f_l = f_c \cdot 2^{\frac{-1}{2N}}

        See also :func:`lower_frequency`.
        """
        return lower_frequency(self.center, self.fraction)

    @property
    def upper(self):
        r"""
        Upper frequency limits of bands.

        .. math:: f_u = f_c \cdot 2^{\frac{+1}{2N}}

        See also :func:`upper_frequency`.
        """
        return upper_frequency(self.center, self.fraction)


__all__ = [
    'exact_center_frequency',
    'nominal_center_frequency',
    'lower_frequency',
    'upper_frequency',
    'index_of_frequency',
    'Octave',
    'frequency_of_band',
    'band_of_frequency',  # These three will be deprecated?
]



"""
Generator
=========

The generator module provides signal generators.

The following functions calculate ``N`` samples and return an array containing the samples.

For indefinitely long iteration over the samples, consider using the output of these functions
in :func:`itertools.cycle`.

Noise
*****

Different types of noise are available. The following table lists the color
of noise and how the power and power density change per octave.

====== ===== =============
Color  Power Power density
====== ===== =============
White  +3 dB  0 dB
Pink    0 dB -3 dB
Blue   +6 dB +3 dB
Brown  -3 dB -6 dB
Violet +9 dB +6 dB
====== ===== =============

The colored noise is created by generating pseudo-random numbers using
:func:`np.random.randn` and then multiplying these with a curve tyical for the color.
Afterwards, an inverse DFT is performed using :func:`np.fft.irfft`.
Finally, the noise is normalized using :func:`acoustics.signal.normalize`.

All colors
----------

.. autofunction:: noise
.. autofunction:: noise_generator

Per color
---------

.. autofunction:: white
.. autofunction:: pink
.. autofunction:: blue
.. autofunction:: brown
.. autofunction:: violet


Other
*****

.. autofunction:: heaviside

For related functions, check :mod:`scipy.signal`.


"""
import itertools
import numpy as np

try:
    from pyfftw.interfaces.numpy_fft import irfft  # Performs much better than numpy's fftpack
except ImportError:  # Use monkey-patching np.fft perhaps instead?
    from numpy.fft import irfft  # pylint: disable=ungrouped-imports

from .signal import normalize


def noise(N, color='white', state=None):
    """Noise generator.

    :param N: Amount of samples.
    :param color: Color of noise.
    :param state: State of PRNG.
    :type state: :class:`np.random.RandomState`

    """
    try:
        return _noise_generators[color](N, state)
    except KeyError:
        raise ValueError("Incorrect color.")


def white(N, state=None):
    """
    White noise.

    :param N: Amount of samples.
    :param state: State of PRNG.
    :type state: :class:`np.random.RandomState`

    White noise has a constant power density. It's narrowband spectrum is therefore flat.
    The power in white noise will increase by a factor of two for each octave band,
    and therefore increases with 3 dB per octave.
    """
    state = np.random.RandomState() if state is None else state
    return state.randn(N)


def pink(N, state=None):
    """
    Pink noise.

    :param N: Amount of samples.
    :param state: State of PRNG.
    :type state: :class:`np.random.RandomState`

    Pink noise has equal power in bands that are proportionally wide.
    Power density decreases with 3 dB per octave.

    """
    # This method uses the filter with the following coefficients.
    #b = np.array([0.049922035, -0.095993537, 0.050612699, -0.004408786])
    #a = np.array([1, -2.494956002, 2.017265875, -0.522189400])
    #return lfilter(B, A, np.random.randn(N))
    # Another way would be using the FFT
    #x = np.random.randn(N)
    #X = rfft(x) / N
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = state.randn(N // 2 + 1 + uneven) + 1j * state.randn(N // 2 + 1 + uneven)
    S = np.sqrt(np.arange(len(X)) + 1.)  # +1 to avoid divide by zero
    y = (irfft(X / S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def blue(N, state=None):
    """
    Blue noise.

    :param N: Amount of samples.
    :param state: State of PRNG.
    :type state: :class:`np.random.RandomState`

    Power increases with 6 dB per octave.
    Power density increases with 3 dB per octave.

    """
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = state.randn(N // 2 + 1 + uneven) + 1j * state.randn(N // 2 + 1 + uneven)
    S = np.sqrt(np.arange(len(X)))  # Filter
    y = (irfft(X * S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def brown(N, state=None):
    """
    Violet noise.

    :param N: Amount of samples.
    :param state: State of PRNG.
    :type state: :class:`np.random.RandomState`

    Power decreases with -3 dB per octave.
    Power density decreases with 6 dB per octave.

    """
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = state.randn(N // 2 + 1 + uneven) + 1j * state.randn(N // 2 + 1 + uneven)
    S = (np.arange(len(X)) + 1)  # Filter
    y = (irfft(X / S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def violet(N, state=None):
    """
    Violet noise. Power increases with 6 dB per octave.

    :param N: Amount of samples.
    :param state: State of PRNG.
    :type state: :class:`np.random.RandomState`

    Power increases with +9 dB per octave.
    Power density increases with +6 dB per octave.

    """
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = state.randn(N // 2 + 1 + uneven) + 1j * state.randn(N // 2 + 1 + uneven)
    S = (np.arange(len(X)))  # Filter
    y = (irfft(X * S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


_noise_generators = {
    'white': white,
    'pink': pink,
    'blue': blue,
    'brown': brown,
    'violet': violet,
}


def noise_generator(N=44100, color='white', state=None):
    """Noise generator.

    :param N: Amount of unique samples to generate.
    :param color: Color of noise.

    Generate `N` amount of unique samples and cycle over these samples.

    """
    #yield from itertools.cycle(noise(N, color)) # Python 3.3
    for sample in itertools.cycle(noise(N, color, state)):
        yield sample


def heaviside(N):
    """Heaviside.

    Returns the value 0 for `x < 0`, 1 for `x > 0`, and 1/2 for `x = 0`.
    """
    return 0.5 * (np.sign(N) + 1)


__all__ = ['noise', 'white', 'pink', 'blue', 'brown', 'violet', 'noise_generator', 'heaviside']



import numpy as np
from numpy.testing import (assert_almost_equal, assert_array_equal, assert_array_almost_equal)

from acoustics.utils import esum, _is_1d, mean_tl, wavelength, w


def test_mean_tl_1d():
    tl = np.array([20, 30, 40, 50])
    surfaces = np.array([50, 40, 30, 20])
    tl_real = 24.1116827
    calculated = mean_tl(tl, surfaces)
    assert_almost_equal(calculated, tl_real)


def test_mean_tl_2d():
    tl = np.array([[20, 30, 40, 50], [20, 30, 40, 50]])
    surfaces = np.array([[50, 40, 30, 20], [1, 10, 11, 22]])
    tl_real = np.array([24.1116827, 33.1466548])
    calculated = mean_tl(tl, surfaces)
    assert_array_almost_equal(calculated, tl_real)


def test_esum_1d():
    calculated = esum(np.array([90, 90, 90]))
    real = 94.77121255
    assert_almost_equal(calculated, real)


def test_esum_2d_default_axis():
    calculated = esum(np.array([[90, 90, 90], [80, 80, 80]]))
    real = np.array(95.18513939877889)
    #real = np.array([94.77121255, 84.77121255])
    assert_array_almost_equal(calculated, real)


def test_esum_2d_axis0():
    calculated = esum(np.array([[90, 90, 90], [80, 80, 80]]), axis=0)
    real = np.array([90.41392685, 90.41392685, 90.41392685])
    assert_almost_equal(calculated, real)


def test__is_1d_float():
    a = 0.9
    is_float = _is_1d(a)
    assert a == is_float


def test__is_1d_1darray():
    a = np.array([0.1, 0.2, 0.3])
    is_1d_array = _is_1d(a)
    a_return = np.array([a])
    assert_array_equal(a_return, is_1d_array)


def test__is_1d_2darray():
    a = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
    is_2d_array = _is_1d(a)
    assert_array_equal(a, is_2d_array)


def test__is1d_vector_2darray():
    a = np.array([[0.1, 0.2, 0.3]])
    is_vector_2darray = _is_1d(a)
    a_return = np.array([0.1, 0.2, 0.3])
    assert_array_equal(a_return, is_vector_2darray)


def test_wavelength1():
    calculated = wavelength(200)
    quasi = 1.715
    assert_almost_equal(calculated, quasi)


def test_wavelength2():
    calculated = wavelength(200, 343)
    quasi = 1.715
    assert_almost_equal(calculated, quasi)


def test_wavelength3():
    freqs = np.array([555, 3333, 11111])
    calculated = wavelength(freqs)
    quasi = np.array([0.618018018, 0.102910291, 0.030870308])
    assert_array_almost_equal(calculated, quasi)


def test_w1():
    calculated = w(1000)
    quasi = 6283.185307179
    assert_almost_equal(calculated, quasi)


def test_w2():
    calculated = w(np.array([1000, 2000, 3000]))
    quasi = np.array([6283.185307179, 12566.370614359, 18849.555921538])
    assert_array_almost_equal(calculated, quasi)



"""
Reflection
==========

The reflection module contains functions for calculating reflection factors and impedances.
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.special import erfc  # pylint: disable=no-name-in-module

SPECIFIC_HEAT_RATIO = 1.4
r"""Specific heat ratio of air :math:`\gamma`."""
POROSITY_DECREASE = 120.0
r"""Rate of exponential decrease of porosity with depth :math:`\alpha`."""
SOUNDSPEED = 343.0
"""Speed of sound in air :math:`c`."""
DENSITY = 1.296
r"""Density of air :math:`\rho`."""


class Boundary:
    """
    An object describing a boundary.
    """

    def __init__(  # pylint: disable=too-many-instance-attributes
            self,
            frequency,
            flow_resistivity,
            density=DENSITY,
            soundspeed=SOUNDSPEED,
            porosity_decrease=POROSITY_DECREASE,
            specific_heat_ratio=SPECIFIC_HEAT_RATIO,
            angle=None,
            distance=None,
            impedance_model='db',
            reflection_model='plane',
    ):

        self.frequency = frequency
        """
        Frequency. Single value or vector for a frequency range.
        """

        self.flow_resistivity = flow_resistivity
        r"""
        Flow resistivity :math:`\sigma`.
        """

        self.density = density
        r"""
        Density of air :math:`\rho`.

        .. note::

            This value is only required for when calculating the impedance according to Attenborough's model.
            See :func:`impedance_attenborough`.

        """

        self.soundspeed = soundspeed
        """
        Speed of sound in air :math:`c`.

        .. note::

            This value is required when calculating the impedance according to Attenborough's model or when
            calculating the spherical wave reflection factor. See respectively :func:`impedance_attenborough`
            and :func:`reflection_factor_spherical_wave`.

        """

        self.porosity_decrease = porosity_decrease
        r"""
        Rate of exponential decrease of porosity with depth :math:`\alpha`.

        .. note::

            This value is only required for when calculating the impedance according to Attenborough's model.
            See :func:`impedance_attenborough`.

        """

        self.specific_heat_ratio = specific_heat_ratio
        r"""
        Ratio of specific heats :math:`\gamma` for air.

        .. note::

            This value is only required for when calculating the impedance according to Attenborough's model.
            See :func:`impedance_attenborough`.

        """

        self.angle = angle
        r"""
        Angle of incidence :math:`\theta`.

        """

        self.distance = distance
        """
        Path length of the reflected ray :math:`r`.

        .. note::

            This value is only required for when calculating the spherical wave reflection factor.
            See :func:`reflection_factor_spherical_wave`.

        """

        self.impedance_model = impedance_model
        """
        Impedance model.

        Possibilities are ``'db'`` and ``'att'`` for respectively :func:`impedance_delany_and_bazley` and :func:`impedance_attenborough`.
        """

        self.reflection_model = reflection_model
        """
        Reflection factor model.

        Possibilities are ``'plane'`` and ``'spherical'``` for respectively :func:`reflection_factor_plane_wave` and :func:`reflection_factor_spherical_wave`.
        """

    @property
    def wavenumber(self):
        r"""
        Wavenumber :math:`k`.

        .. math:: k = \frac{2 \pi f}{c}
        """
        return 2.0 * np.pi * self.frequency / self.soundspeed

    @property
    def impedance(self):
        """
        Impedance according to chosen impedance model defined using :meth:`impedance_model`.
        """
        if self.impedance_model == 'db':
            return impedance_delany_and_bazley(self.frequency, self.flow_resistivity)
        if self.impedance_model == 'att':
            return impedance_attenborough(
                self.frequency,
                self.flow_resistivity,
                self.density,
                self.soundspeed,
                self.porosity_decrease,
                self.specific_heat_ratio,
            )
        else:
            raise ValueError("Incorrect impedance model.")

    @property
    def reflection_factor(self):
        """
        Reflection factor according to chosen reflection factor model defined using :meth:`reflection_model`.
        """
        if self.angle is None:
            raise AttributeError('Cannot calculate reflection factor. self.angle has not been specified.')

        if self.reflection_model == 'plane':
            return reflection_factor_plane_wave(*np.meshgrid(self.impedance, self.angle))
        elif self.reflection_model == 'spherical':
            if self.distance is None:
                raise AttributeError('Cannot calculate reflection factor. self.distance has not been specified.')
            else:
                return reflection_factor_spherical_wave(
                    *np.meshgrid(self.impedance, self.angle),
                    distance=self.distance,
                    wavenumber=self.wavenumber,
                )
        else:
            raise RuntimeError("Oops...")

    def plot_impedance(self, filename=None):
        """
        Plot magnitude and phase of the impedance as function of frequency.
        """

        fig = plt.figure()

        ax0 = fig.add_subplot(211)
        ax0.set_title('Magnitude of impedance')
        ax0.semilogx(self.frequency, np.abs(self.impedance))
        ax0.set_xlabel(r'$f$ in Hz')
        ax0.set_ylabel(r'$\left|Z\right|$')
        ax0.grid()

        ax0 = fig.add_subplot(212)
        ax0.set_title('Angle of impedance')
        ax0.semilogx(self.frequency, np.angle(self.impedance))
        ax0.set_xlabel(r'$f$ in Hz')
        ax0.set_ylabel(r'$\angle Z$')
        ax0.grid()

        plt.tight_layout()

        if filename:
            fig.savefig(filename, transparant=True)
        return fig

    def plot_reflection_factor(self, filename=None):
        """
        Plot reflection factor.
        """

        if self.frequency is None:
            raise ValueError("No frequency specified.")
        if self.angle is None:
            raise ValueError("No angle specified.")

        try:
            n_f = len(self.frequency)
        except TypeError:
            n_f = 1
        try:
            n_a = len(self.angle)
        except TypeError:
            n_a = 1

        if n_f == 1 and n_a == 1:
            raise ValueError("Either frequency or angle needs to be a vector.")

        elif n_f == 1 or n_a == 1:
            if n_f == 1 and n_a > 1:  # Show R as function of angle for a single frequency.
                xlabel = r"$\theta$ in degrees"
            elif n_f > 1 and n_a == 1:  # Show R as function of frequency for a single angle.
                xlabel = r"$f$ in Hz"
            R = self.reflection_factor
            fig = plt.figure()

            ax0 = fig.add_subplot(211)
            ax0.set_title("Magnitude of reflection factor")
            ax0.semilogx(self.frequency, np.abs(R))
            ax0.set_xlabel(xlabel)
            ax0.set_ylabel(r'$\left|R\right|$')
            ax0.grid()

            ax1 = fig.add_subplot(212)
            ax1.set_title("Phase of reflection factor")
            ax1.semilogx(self.frequency, np.angle(R))
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(r'$\angle R$')
            ax1.grid()

        elif n_f > 1 and n_a > 1:  # Show 3D or pcolor
            R = self.reflection_factor
            fig = plt.figure()

            #grid = AxesGrid(fig, 111, nrows_ncols=(2, 2), axes_pad=0.1, cbar_mode='each', cbar_location='right')
            ax0 = fig.add_subplot(211)
            #ax0 = grid[0]
            ax0.set_title("Magnitude of reflection factor")
            ax0.pcolormesh(self.frequency, self.angle * 180.0 / np.pi, np.abs(R))
            #ax0.pcolor(self.angle, self.frequency, np.abs(R))
            #ax0.set_xlabel(xlabel)
            #ax0.set_ylabel(r'$\left|R\right|$')
            ax0.grid()

            ax1 = fig.add_subplot(212)
            #ax1 = grid[1]
            ax1.set_title("Phase of reflection factor")
            ax1.pcolormesh(self.frequency, self.angle * 180.0 / np.pi, np.angle(R))
            #ax1.pcolor(self.angle, self.frequency, np.angle(R))
            #ax0.set_xlabel(xlabel)
            #ax0.set_ylabel(r'$\angle R$')
            ax1.grid()

        else:
            raise RuntimeError("Oops...")

        #plt.tight_layout()

        if filename:
            fig.savefig(filename, transparant=True)
        else:
            return fig


def reflection_factor_plane_wave(impedance, angle):
    r"""
    Plane wave reflection factor :math:`R`.

    :param impedance: Normalized impedance :math:`Z`.
    :param angle: Angle of incidence :math:`\theta`.


    The plane wave reflection factor :math:`R` is given by

    .. math:: R = \frac{Z\cos{\theta}-1}{Z\cos{\theta}+1}

    where :math:`Z` is the normalized impedance and :math:`\theta` the angle of incidence.

    """
    return (impedance * np.cos(angle) - 1.0) / (impedance * np.cos(angle) + 1.0)


def numerical_distance(impedance, angle, distance, wavenumber):
    r"""
    Numerical distance :math:`w`.

    :param impedance: Normalized impedance :math:`Z`.
    :param angle: Angle of incidence :math:`\theta`.
    :param distance: Path length of the reflected ray :math:`r`.
    :param wavenumber: Wavenumber :math:`k`.

    The numerical distance :math:`w` is given by

    .. math:: w = \sqrt{-j k r  \left( 1 + \frac{1}{Z} \cos{\theta} - \sqrt{1 - \left( \frac{1}{Z} \right)^2} \sin{\theta} \right) }

    """
    return np.sqrt(-1j * wavenumber * distance *
                   (1.0 + 1.0 / impedance * np.cos(angle) - np.sqrt(1.0 - (1.0 / impedance)**2.0) * np.sin(angle)))


def reflection_factor_spherical_wave(impedance, angle, distance, wavenumber):
    r"""
    Spherical wave reflection factor :math:`Q`.

    :param impedance: Normalized impedance :math:`Z`.
    :param angle: Angle of incidence :math:`\theta`.
    :param distance: Path length of the reflected ray :math:`r`.
    :param wavenumber: Wavenumber :math:`k`.

    The spherical wave relfection factor :math:`Q` is given by

    .. math:: Q = R \left(1 - R \right) F

    where :math:`R` is the plane wave reflection factor as calculated in :func:`reflection_factor_plane_wave` and :math:`F` is given by

    .. math:: F = 1 - j \sqrt{ \pi} w e^{-w^2} \mathrm{erfc} \left( j w \right)

    where :math:`w` is the numerical distance as calculated in :func:`numerical_distance`.
    """
    w = numerical_distance(impedance, angle, distance, wavenumber)

    F = 1.0 - 1j * np.sqrt(np.pi) * w * np.exp(-w**2.0) * erfc(1j * w)

    plane_factor = reflection_factor_plane_wave(impedance, angle)
    return plane_factor * (1.0 - plane_factor) * F


def impedance_delany_and_bazley(frequency, flow_resistivity):
    r"""
    Normalised specific acoustic impedance according to the empirical one-parameter model by Delany and Bazley.

    :param frequency: Frequency :math:`f`.
    :param flow_resistivity: Flow resistivity :math:`\sigma`.

    The impedance :math:`Z` is given by

    .. math:: Z = 1 + 9.08 \left( \frac{1000f}{\sigma}\right)^{-0.75} - 11.9 j \left( \frac{1000f}{\sigma}\right)^{-0.73}

    """
    return 1.0 + 9.08 * (1000.0 * frequency / flow_resistivity)**(-0.75) - 1j * 11.9 * (
        1000.0 * frequency / flow_resistivity)**(-0.73)


def impedance_attenborough(
        frequency,
        flow_resistivity,
        density=DENSITY,
        soundspeed=SOUNDSPEED,
        porosity_decrease=POROSITY_DECREASE,
        specific_heat_ratio=SPECIFIC_HEAT_RATIO,
):
    r"""
    Normalised specific acoustics impedance according to the two-parameter model by Attenborough.

    :param frequency: Frequency :math:`f`.
    :param flow_resistivity: Flow resistivity :math:`\sigma`.
    :param soundspeed: Speed of sound in air :math:`c`.
    :param density: Density of air :math:`\rho`.
    :param porosity_decrease: Rate of exponential decrease of porosity with depth :math:`\alpha`.
    :param specific_heat_ratio: Ratio of specific heats :math:`\gamma` for air.


    The impedance :math:`Z` is given by

    .. math:: Z = \frac{\left( 1-j\right) \sqrt{\sigma/f}}{\sqrt{\pi \gamma_0 \rho_0}} - \frac{jc\alpha}{8 \pi \gamma_0 f}

    """
    return (1.0 - 1j) * np.sqrt(flow_resistivity / frequency) / np.sqrt(
        np.pi * specific_heat_ratio * density) - 1j * soundspeed * porosity_decrease / (
            8.0 * np.pi * specific_heat_ratio * frequency)



import numpy as np

#from scipy.signal import (complex_cepstrum, inverse_complex_cepstrum,
#                          real_cepstrum, minimum_phase, sawtooth)
from scipy.signal import sawtooth
from acoustics.cepstrum import *

from numpy.testing import (assert_array_almost_equal)


def test_complex_cepstrum():
    """The period of a periodic harmonic will show up as a peak in a
    complex cepstrum.
    """
    duration = 5.0
    fs = 8000.0
    samples = int(fs * duration)
    t = np.arange(samples) / fs
    fundamental = 100.0
    signal = sawtooth(2. * np.pi * fundamental * t)
    ceps, _ = complex_cepstrum(signal)
    assert (fundamental == 1.0 / t[ceps.argmax()])


def test_inverse_complex_cepstrum():
    """Applying the complex cepstrum and then the inverse complex cepstrum
    should result in the original sequence.
    """
    x = np.arange(10)
    ceps, ndelay = complex_cepstrum(x)
    y = inverse_complex_cepstrum(ceps, ndelay)
    assert_array_almost_equal(x, y)



import os
import itertools

import pytest
import numpy as np

from get_data_path import data_path

details_type = ["det_global", "det_oct", "det_third"]
int_time = ["2000ms", "1000ms", "0500ms", "0250ms", "0063ms", "0010ms"]
delimiter = ["comma", "semicolon"]
decimals = ["1dec", "2dec"]
time_history = ["overall", "timehist"]
weighting = ["z", "za", "zc", "zac"]
cirrus_files = list(itertools.product(
    details_type,
    int_time,
    delimiter,
    decimals,
))
cirrus_files.extend(list(itertools.product(
    ["det_global_general"],
    delimiter,
    decimals,
)))
cirrus_files.extend(list(itertools.product(
    time_history,
    ["oct"],
    weighting,
    delimiter,
    decimals,
)))
cirrus_files.extend(list(itertools.product(
    ["overall_third"],
    weighting,
    delimiter,
    decimals,
)))


@pytest.mark.parametrize('filename', cirrus_files)
def test_read_csv_cirrus_details(filename):
    from acoustics.aio import read_csv_cirrus
    file = "_".join(filename) + ".csv"
    csv_path = os.path.join(data_path(), "cirrus", file)
    data = read_csv_cirrus(csv_path)
    if filename[0] == "det_global":
        np.sum(data.LAeq)
    else:
        np.sum(data["125Hz"])



import pytest
import numpy as np
from acoustics.standards.iso_1996_2_2007 import Tonality
from acoustics import Signal
import collections


def test_tonality():

    duration = 60.0
    fs = 10025.0
    samples = int(fs * duration)
    times = np.arange(samples) / fs

    signal = Signal(np.sin(2.0 * np.pi * 1000.0 * times), fs)

    tonality = Tonality(signal, signal.fs)

    # Test methods before analysis
    tonality.spectrum
    tonality.plot_spectrum()

    tonality.frequency_resolution
    tonality.effective_analysis_bandwidth

    # No values yet, cannot print overview.
    with pytest.raises(ValueError):
        print(tonality.overview())
    tonality.results_as_dataframe()

    assert len(list(tonality.noise_pauses)) == 0
    assert len(list(tonality.tones)) == 0
    assert len(list(tonality.critical_bands)) == 0

    # Perform analysis
    tonality.determine_noise_pauses().analyse()

    # Needs to be checked
    #assert len(list(tonality.noise_pauses)) == 1
    #assert len(list(tonality.tones)) == 1
    #assert len(list(tonality.critical_bands)) == 1

    tonality.critical_band_at(900.0)

    tonality.dominant_tone
    print(tonality.overview())
    tonality.results_as_dataframe()
    tonality.plot_results()



#Target = collections.namedtuple('Target', ['tonal_level',
                                           #'masking_noise_level',
                                           #'tonal_audibility',
                                           #'adjustment'])
#TARGET_DATA = {
    #'example1': Target(46.7, 37.3, 13.7, 6.0),
    #'example2': Target(54.1, 45.2, 11.1, 6.0),
    #'example3': Target(54.6, 45.5, 10.6, 6.0),
    ##'example4': Target(53.6, 45.5, 10.7, 6.0), # Time-varying test.
    #}

#@pytest.fixture(params=TARGET_DATA.keys())
#def example(request):
    #return request.param

#def test_verify_standard_examples(example):
    #"""Verify against the examples shown in the standard.


    #.. note:: The FTP server is slow. Be patient.

    #For the examples in the standard:
    #- Number of spectra: 350
    #- Signal duration: 2 minutes
    #- Window: Hanning
    #- Averaging: Linear
    #- Effective analysis bandwidth: 4.39 Hz

    #The sample frequency is 44.1 kHz. Dividing this sample frequency by
    #results in 360 spectra and an effective analysis bandwidth of
    #4.5 Hz, which is very close to the example.

    #"""
    #from ftplib import FTP
    #import tempfile
    #import os

    #FOLDER = 'THP wave files/'
    #FILES = {'calibration' : 'cal. 93,8 dB.wav',
             #'example1': 'sam 03 eks 1.wav',
             #'example2': 'sam 03 eks 2.wav',
             #'example3': 'sam 03 eks 3.wav',
             #'example4': 'sam 03 eks 4.wav',
             #}
    #CHANNEL = 0

    #fs = 44100.0
    #duration = 120.0
    #samples = duration*fs
    #spectra = 350
    #bins = samples // spectra

    ## Obtain data from Delta
    ## https://noiselabdk.wordpress.com/demo-download/
    ##hostname = 'ftp.delta.dk'
    ##remote_data = 'Demo Files/Tone wave files.zip'
    ##username = 'nlpublic'
    ##password = 'noiselab'
    ##local_data = 'data.zip'

    ##with tempfile.TemporaryDirectory() as directory:
        ##ftp = FTP(hostname, username, password)
        ##with open(local_data, 'wb') as target_file:
            ##ftp.retrbinary('RETR %s' % remote_data, target_file.write)



    #calibration_at = 93.8 # decibel
    #signal = Signal.from_wav(os.path.join(FOLDER, FILES['calibration']))[CHANNEL]
    #calibration_factor = 10.0**((calibration_at - signal.leq())/20.0)

    #target = TARGET_DATA[example]
    #filename = os.path.join(FOLDER, FILES[example])

    #signal = Signal.from_wav(filename)[CHANNEL].pick(0.0, 60.0).weigh('A') * calibration_factor
    #tonality = Tonality(signal, signal.fs, bins=bins)
    #tonality.determine_noise_pauses().analyse()

    #cb = tonality.dominant_tone.critical_band
    #print(tonality.results())
    #assert abs(tonality.dominant_tone.tone_level - target.tonal_level) < 1.0
    #assert abs(cb.masking_noise_level - target.masking_noise_level) < 1.0
    #assert abs(cb.tonal_audibility - target.tonal_audibility) < 1.0
    #assert abs(cb.adjustment - target.adjustment) < 1.0



import itertools
import matplotlib.pyplot as plt
import numpy as np
from scipy.io import wavfile
from scipy.signal import detrend, lfilter, bilinear, spectrogram, filtfilt, resample, fftconvolve
import acoustics

from acoustics.standards.iso_tr_25417_2007 import REFERENCE_PRESSURE
from acoustics.standards.iec_61672_1_2013 import WEIGHTING_SYSTEMS
from acoustics.standards.iec_61672_1_2013 import (NOMINAL_OCTAVE_CENTER_FREQUENCIES,
                                                  NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES)


class Signal(np.ndarray):
    """A signal consisting of samples (array) and a sample frequency (float).

    """

    def __new__(cls, data, fs):
        obj = np.asarray(data).view(cls)
        obj.fs = fs
        return obj

    def __array_prepare__(self, array, context=None):
        try:
            a = context[1][0]
            b = context[1][1]
        except IndexError:
            return array

        if hasattr(a, 'fs') and hasattr(b, 'fs'):
            if a.fs == b.fs:
                return array
            else:
                raise ValueError("Sample frequencies do not match.")
        else:
            return array

    def __array_wrap__(self, out_arr, context=None):
        return np.ndarray.__array_wrap__(self, out_arr, context)

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return

        self.fs = getattr(obj, 'fs', None)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(Signal, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (self.fs, )
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.fs = state[-1]  # Set the info attribute
        # Call the parent's __setstate__ with the other tuple elements.
        super(Signal, self).__setstate__(state[0:-1])

    def __repr__(self):
        return "Signal({})".format(str(self))

    def _construct(self, x):
        """Construct signal like x."""
        return Signal(x, self.fs)

    @property
    def samples(self):
        """Amount of samples in signal."""
        return self.shape[-1]

    @property
    def channels(self):
        """Amount of channels.
        """
        if self.ndim > 1:
            return self.shape[-2]
        else:
            return 1

    @property
    def duration(self):
        """Duration of signal in seconds.
        """
        return float(self.samples / self.fs)

    @property
    def values(self):
        """Return the values of this signal as an instance of :class:`np.ndarray`."""
        return np.array(self)

    def calibrate_to(self, decibel, inplace=False):
        """Calibrate signal to value `decibel`.

        :param decibel: Value to calibrate to.
        :param inplace: Whether to perform inplace or not.
        :returns: Calibrated signal.
        :rtype: :class:`Signal`

        Values of `decibel` are broadcasted. To set a value per channel, use `decibel[...,None]`.
        """
        decibel = decibel * np.ones(self.shape)
        gain = decibel - self.leq()[..., None]
        return self.gain(gain, inplace=inplace)

    def calibrate_with(self, other, decibel, inplace=False):
        """Calibrate signal with other signal.

        :param other: Other signal/array.
        :param decibel: Signal level of `other`.
        :param inplace: Whether to perform inplace or not.
        :returns: calibrated signal.
        :rtype: :class:`Signal`
        """
        if not isinstance(other, Signal):
            other = Signal(other, self.fs)
        gain = decibel - other.leq()
        return self.gain(gain, inplace=inplace)

    def decimate(self, factor, zero_phase=False, ftype='iir', order=None):
        """Decimate signal by integer `factor`. Before downsampling a low-pass filter is applied.

        :param factor: Downsampling factor.
        :param zero_phase: Prevent phase shift by filtering with ``filtfilt`` instead of ``lfilter``.
        :param ftype: Filter type.
        :param order: Filter order.
        :returns: Decimated signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`scipy.signal.decimate`
        .. seealso:: :meth:`resample`

        """
        return Signal(
            acoustics.signal.decimate(x=self, q=factor, n=order, ftype=ftype, zero_phase=zero_phase), self.fs / factor)

    def resample(self, nsamples, times=None, axis=-1, window=None):
        """Resample signal.

        :param samples: New amount of samples.
        :param times: Times corresponding to samples.
        :param axis: Axis.
        :param window: Window.

        .. seealso:: :func:`scipy.signal.resample`
        .. seealso:: :meth:`decimate`

        You might want to low-pass filter this signal before resampling.

        """
        return Signal(resample(self, nsamples, times, axis, window), nsamples / self.samples * self.fs)

    def upsample(self, factor, axis=-1):
        """Upsample signal with integer factor.

        :param factor: Upsample factor.
        :param axis: Axis.

        .. seealso:: :meth:`resample`
        """
        return self.resample(int(self.samples * factor), axis=axis)

    def gain(self, decibel, inplace=False):
        """Apply gain of `decibel` decibels.

        :param decibel: Decibels
        :param inplace: In place
        :returns: Amplified signal.
        :rtype: :class:`Signal`
        """
        factor = 10.0**(decibel / 20.0)
        if inplace:
            self *= factor
            return self
        else:
            return self * factor

    def pick(self, start=0.0, stop=None):
        """Get signal from start time to stop time.

        :param start: Start time.
        :type start: float
        :param stop: End time.
        :type stop: float
        :returns: Selected part of the signal.
        :rtype: :class:`Signal`

        """
        if start is not None:
            start = int(np.floor(start * self.fs))
        if stop is not None:
            stop = int(np.floor(stop * self.fs))
        return self[..., start:stop]

    def times(self):
        """Time vector.

        :returns: A vector with a timestamp for each sample.
        :rtype: :class:`np.ndarray`

        """
        return np.arange(0, self.samples) / self.fs

    def energy(self):
        """Signal energy.

        :returns: Total energy per channel.
        :rtype: :class:`np.ndarray`

        .. math:: E = \\sum_{n=0}^{N-1} |x_n|^2

        """
        return float((self * self).sum())

    def power(self):
        """Signal power.

        .. math:: P = \\frac{1}{N} \\sum_{n=0}^{N-1} |x_n|^2
        """
        return self.energy() / len(self)

    def ms(self):
        """Mean value squared of signal.

        .. seealso:: :func:`acoustics.signal.ms`

        """
        return acoustics.signal.ms(self)

    def rms(self):
        """Root mean squared of signal.

        .. seealso:: :func:`acoustics.signal.rms`

        """
        return acoustics.signal.rms(self)
        #return np.sqrt(self.power())

    def weigh(self, weighting='A', zero_phase=False):
        """Apply frequency-weighting. By default 'A'-weighting is applied.

        :param weighting: Frequency-weighting filter to apply.
            Valid options are 'A', 'C' and 'Z'. Default weighting is 'A'.
        :returns: Weighted signal.
        :rtype: :class:`Signal`.

        By default the weighting filter is applied using
        :func:`scipy.signal.lfilter` causing a frequency-dependent delay. In case a
        delay is undesired, the filter can be applied using :func:`scipy.signal.filtfilt`
        by setting `zero_phase=True`.

        """
        num, den = WEIGHTING_SYSTEMS[weighting]()
        b, a = bilinear(num, den, self.fs)
        func = filtfilt if zero_phase else lfilter
        return self._construct(func(b, a, self))

    def correlate(self, other=None, mode='full'):
        """Correlate signal with `other` signal. In case `other==None` this
        method returns the autocorrelation.

        :param other: Other signal.
        :param mode: Mode.

        .. seealso:: :func:`np.correlate`, :func:`scipy.signal.fftconvolve`

        """
        if other is None:
            other = self
        if self.fs != other.fs:
            raise ValueError("Cannot correlate. Sample frequencies are not the same.")
        if self.channels > 1 or other.channels > 1:
            raise ValueError("Cannot correlate. Not supported for multichannel signals.")
        return self._construct(fftconvolve(self, other[::-1], mode=mode))

    def amplitude_envelope(self):
        """Amplitude envelope.

        :returns: Amplitude envelope of signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`acoustics.signal.amplitude_envelope`

        """
        return self._construct(acoustics.signal.amplitude_envelope(self, self.fs))

    def instantaneous_frequency(self):
        """Instantaneous frequency.

        :returns: Instantaneous frequency of signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`acoustics.signal.instantaneous_frequency`

        """
        return self._construct(acoustics.signal.instantaneous_frequency(self, self.fs))

    def instantaneous_phase(self):
        """Instantaneous phase.

        :returns: Instantaneous phase of signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`acoustics.signal.instantaneous_phase`

        """
        return self._construct(acoustics.signal.instantaneous_phase(self, self.fs))

    def detrend(self, **kwargs):
        """Detrend signal.

        :returns: Detrended version of signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`scipy.signal.detrend`

        """
        return self._construct(detrend(self, **kwargs))

    def unwrap(self):
        """Unwrap signal in case the signal represents wrapped phase.

        :returns: Unwrapped signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`np.unwrap`

        """
        return self._construct(np.unwrap(self))

    def complex_cepstrum(self, N=None):
        """Complex cepstrum.

        :param N: Amount of bins.
        :returns: Quefrency, complex cepstrum and delay in amount of samples.

        .. seealso:: :func:`acoustics.cepstrum.complex_cepstrum`

        """
        if N is not None:
            times = np.linspace(0.0, self.duration, N, endpoint=False)
        else:
            times = self.times()
        cepstrum, ndelay = acoustics.cepstrum.complex_cepstrum(self, n=N)
        return times, cepstrum, ndelay

    def real_cepstrum(self, N=None):
        """Real cepstrum.

        :param N: Amount of bins.
        :returns: Quefrency and real cepstrum.

        .. seealso:: :func:`acoustics.cepstrum.real_cepstrum`

        """
        if N is not None:
            times = np.linspace(0.0, self.duration, N, endpoint=False)
        else:
            times = self.times()
        return times, acoustics.cepstrum.real_cepstrum(self, n=N)

    def power_spectrum(self, N=None):
        """Power spectrum.

        :param N: Amount of bins.

        .. seealso:: :func:`acoustics.signal.power_spectrum`

        """
        return acoustics.signal.power_spectrum(self, self.fs, N=N)

    def angle_spectrum(self, N=None):
        """Phase angle spectrum. Wrapped.

        :param N: amount of bins.

        .. seealso::

            :func:`acoustics.signal.angle_spectrum`, :func:`acoustics.signal.phase_spectrum`
            and :meth:`phase_spectrum`.

        """
        return acoustics.signal.angle_spectrum(self, self.fs, N=N)

    def phase_spectrum(self, N=None):
        """Phase spectrum. Unwrapped.

        :param N: Amount of bins.

        .. seealso::

            :func:`acoustics.signal.phase_spectrum`, :func:`acoustics.signal.angle_spectrum`
            and :meth:`angle_spectrum`.

        """
        return acoustics.signal.phase_spectrum(self, self.fs, N=N)

    def peak(self, axis=-1):
        """Peak sound pressure.

        :param axis: Axis.

        .. seealso::

            :func:`acoustic.standards.iso_tr_25417_2007.peak_sound_pressure`

        """
        return acoustics.standards.iso_tr_25417_2007.peak_sound_pressure(self, axis=axis)

    def peak_level(self, axis=-1):
        """Peak sound pressure level.

        :param axis: Axis.

        .. seealso::

            :func:`acoustics.standards.iso_tr_25417_2007.peak_sound_pressure_level`

        """
        return acoustics.standards.iso_tr_25417_2007.peak_sound_pressure_level(self, axis=axis)

    def min(self, axis=-1):
        """Return the minimum along a given axis.

        Refer to `np.amin` for full documentation.
        """
        return np.ndarray.min(self, axis=axis)

    def max(self, axis=-1):
        """Return the minimum along a given axis.

        Refer to `np.amax` for full documentation.
        """
        return np.ndarray.max(self, axis=axis)

    def max_level(self, axis=-1):
        """Maximum sound pressure level.

        :param axis: Axis.

        .. seealso:: :func:`acoustics.standards.iso_tr_25417_2007.max_sound_pressure_level`

        """
        return acoustics.standards.iso_tr_25417_2007.max_sound_pressure_level(self, axis=axis)

    def sound_exposure(self, axis=-1):
        """Sound exposure.

        :param axis: Axis.

        .. seealso:: :func:`acoustics.standards.iso_tr_25417_2007.sound_exposure`

        """
        return acoustics.standards.iso_tr_25417_2007.sound_exposure(self, self.fs, axis=axis)

    def sound_exposure_level(self, axis=-1):
        """Sound exposure level.

        :param axis: Axis.

        .. seealso:: :func:`acoustics.standards.iso_tr_25417_2007.sound_exposure_level`

        """
        return acoustics.standards.iso_tr_25417_2007.sound_exposure_level(self, self.fs, axis=axis)

    def plot_complex_cepstrum(self, N=None, **kwargs):
        """Plot complex cepstrum of signal.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * frequency: Boolean indicating whether the x-axis should show time in seconds or quefrency
        * xlabel_frequency: Label in case frequency is shown.

        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': "$t$ in s",
            'ylabel': "$C$",
            'title': 'Complex cepstrum',
            'frequency': False,
            'xlabel_frequency': "$f$ in Hz",
        }
        params.update(kwargs)

        t, ceps, _ = self.complex_cepstrum(N=N)
        if params['frequency']:
            t = 1. / t
            params['xlabel'] = params['xlabel_frequency']
            t = t[::-1]
            ceps = ceps[::-1]
        return _base_plot(t, ceps, params)

    def plot_real_cepstrum(self, N=None, **kwargs):
        """Plot real cepstrum of signal.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * frequency: Boolean indicating whether the x-axis should show time in seconds or quefrency
        * xlabel_frequency: Label in case frequency is shown.

        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': "$t$ in s",
            'ylabel': "$C$",
            'title': 'Real cepstrum',
            'frequency': False,
            'xlabel_frequency': "$f$ in Hz",
        }
        params.update(kwargs)

        t, ceps = self.real_cepstrum(N=N)
        if params['frequency']:
            t = 1. / t
            params['xlabel'] = params['xlabel_frequency']
            t = t[::-1]
            ceps = ceps[::-1]
        return _base_plot(t, ceps, params)

    def plot_power_spectrum(self, N=None, **kwargs):  #filename=None, scale='log'):
        """Plot spectrum of signal.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * reference: Reference power

        .. seealso:: :meth:`power_spectrum`

        """
        params = {
            'xscale': 'log',
            'yscale': 'linear',
            'xlabel': "$f$ in Hz",
            'ylabel': "$L_{p}$ in dB",
            'title': 'SPL',
            'reference': REFERENCE_PRESSURE**2.0,
        }
        params.update(kwargs)

        f, o = self.power_spectrum(N=N)
        return _base_plot(f, 10.0 * np.log10(o / params['reference']), params)

    def plot_angle_spectrum(self, N=None, **kwargs):
        """Plot phase angle spectrum of signal. Wrapped.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * reference: Reference power

        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': "$f$ in Hz",
            'ylabel': r"$\angle \phi$",
            'title': 'Phase response (wrapped)',
        }
        params.update(kwargs)
        f, o = self.angle_spectrum(N=N)
        return _base_plot(f, o, params)

    def plot_phase_spectrum(self, N=None, **kwargs):
        """Plot phase spectrum of signal. Unwrapped.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * reference: Reference power

        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': "$f$ in Hz",
            'ylabel': r"$\angle \phi$",
            'title': 'Phase response (unwrapped)',
        }
        params.update(kwargs)
        f, o = self.phase_spectrum(N=N)
        return _base_plot(f, o, params)

    def spectrogram(self, **kwargs):
        """Spectrogram of signal.

        :returns: Spectrogram.

        See :func:`scipy.signal.spectrogram`. Some of the default values have been changed.
        The generated spectrogram consists by default of complex values.

        """
        params = {
            'nfft': 4096,
            'noverlap': 128,
            'mode': 'complex',
        }
        params.update(kwargs)

        t, s, P = spectrogram(self, fs=self.fs, **params)

        return t, s, P

    def plot_spectrogram(self, **kwargs):
        """
        Plot spectrogram of the signal.

        Valid kwargs:

        * xlim
        * ylim
        * clim
        .. note:: This method only works for a single channel.

        """
        # To do, use :meth:`spectrogram`.
        params = {
            'xlim': None,
            'ylim': None,
            'clim': None,
            'NFFT': 4096,
            'noverlap': 128,
            'title': 'Spectrogram',
            'xlabel': '$t$ in s',
            'ylabel': '$f$ in Hz',
            'clabel': 'SPL in dB',
            'colorbar': True,
        }
        params.update(kwargs)

        if self.channels > 1:
            raise ValueError("Cannot plot spectrogram of multichannel signal. Please select a single channel.")

        # Check if an axes object is passed in. Otherwise, create one.
        ax0 = params.get('ax', plt.figure().add_subplot(111))
        ax0.set_title(params['title'])

        data = np.squeeze(self)
        try:
            _, _, _, im = ax0.specgram(data, Fs=self.fs, noverlap=params['noverlap'], NFFT=params['NFFT'],
                                       mode='magnitude', scale_by_freq=False)
        except AttributeError:
            raise NotImplementedError(
                "Your version of matplotlib is incompatible due to lack of support of the mode keyword argument to matplotlib.mlab.specgram."
            )

        if params['colorbar']:
            cb = ax0.get_figure().colorbar(mappable=im)
            cb.set_label(params['clabel'])

        ax0.set_xlim(params['xlim'])
        ax0.set_ylim(params['ylim'])
        im.set_clim(params['clim'])

        ax0.set_xlabel(params['xlabel'])
        ax0.set_ylabel(params['ylabel'])

        return ax0

    def levels(self, time=0.125, method='average'):
        """Calculate sound pressure level as function of time.

        :param time: Averaging time or integration time constant. Default value is 0.125 corresponding to FAST.
        :param method: Use time `average` or time `weighting`. Default option is `average`.
        :returns: sound pressure level as function of time.

        .. seealso:: :func:`acoustics.standards.iec_61672_1_2013.time_averaged_sound_level`
        .. seealso:: :func:`acoustics.standards.iec_61672_1_2013.time_weighted_sound_level`

        """
        if method == 'average':
            return acoustics.standards.iec_61672_1_2013.time_averaged_sound_level(self.values, self.fs, time)
        elif method == 'weighting':
            return acoustics.standards.iec_61672_1_2013.time_weighted_sound_level(self.values, self.fs, time)
        else:
            raise ValueError("Invalid method")

    def leq(self):
        """Equivalent level. Single-value number.

        .. seealso:: :func:`acoustics.standards.iso_tr_25417_2007.equivalent_sound_pressure_level`

        """
        return acoustics.standards.iso_tr_25417_2007.equivalent_sound_pressure_level(self.values)

    def plot_levels(self, **kwargs):
        """Plot sound pressure level as function of time.

        .. seealso:: :meth:`levels`

        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': '$t$ in s',
            'ylabel': '$L_{p,F}$ in dB',
            'title': 'SPL',
            'time': 0.125,
            'method': 'average',
            'labels': None,
        }
        params.update(kwargs)
        t, L = self.levels(params['time'], params['method'])
        L_masked = np.ma.masked_where(np.isinf(L), L)
        return _base_plot(t, L_masked, params)

    #def octave(self, frequency, fraction=1):
    #"""Determine fractional-octave `fraction` at `frequency`.

    #.. seealso:: :func:`acoustics.signal.fractional_octaves`

    #"""
    #return acoustics.signal.fractional_octaves(self, self.fs, frequency,
    #frequency, fraction, False)[1]

    def bandpass(self, lowcut, highcut, order=8, zero_phase=False):
        """Filter signal with band-pass filter.

        :param lowcut: Lower cornerfrequency.
        :param highcut: Upper cornerfrequency.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        :returns: Band-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.bandpass`
        """
        return type(self)(acoustics.signal.bandpass(self, lowcut, highcut, self.fs, order=order, zero_phase=zero_phase),
                          self.fs)

    def bandstop(self, lowcut, highcut, order=8, zero_phase=False):
        """Filter signal with band-stop filter.

        :param lowcut: Lower cornerfrequency.
        :param highcut: Upper cornerfrequency.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        :returns: Band-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.bandstop`
        """
        return type(self)(acoustics.signal.bandstop(self, lowcut, highcut, self.fs, order=order, zero_phase=zero_phase),
                          self.fs)

    def highpass(self, cutoff, order=4, zero_phase=False):
        """Filter signal with high-pass filter.

        :param cutoff: Cornerfrequency.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: High-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.highpass`
        """
        return type(self)(acoustics.signal.highpass(self, cutoff, self.fs, order=order, zero_phase=zero_phase), self.fs)

    def lowpass(self, cutoff, order=4, zero_phase=False):
        """Filter signal with low-pass filter.

        :param cutoff: Cornerfrequency.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Low-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.lowpass`
        """
        return type(self)(acoustics.signal.lowpass(self, cutoff, self.fs, order=order, zero_phase=zero_phase), self.fs)

    def octavepass(self, center, fraction, order=8, zero_phase=False):
        """Filter signal with fractional-octave band-pass filter.

        :param center: Center frequency. Any value in the band will suffice.
        :param fraction: Band designator.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Band-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.octavepass`
        """
        return type(self)(acoustics.signal.octavepass(self, center, self.fs, fraction=fraction, order=order,
                                                      zero_phase=zero_phase), self.fs)

    def bandpass_frequencies(self, frequencies, order=8, purge=True, zero_phase=False):
        """Apply bandpass filters for frequencies.

        :param frequencies: Band-pass filter frequencies.
        :type frequencies: Instance of :class:`acoustics.signal.Frequencies`
        :param order: Filter order.
        :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Frequencies and band-pass filtered signal.

        .. seealso:: :func:`acoustics.signal.bandpass_frequencies`
        """
        frequencies, filtered = acoustics.signal.bandpass_frequencies(self, self.fs, frequencies, order, purge,
                                                                      zero_phase=zero_phase)
        return frequencies, type(self)(filtered, self.fs)

    def octaves(self, frequencies=NOMINAL_OCTAVE_CENTER_FREQUENCIES, order=8, purge=True, zero_phase=False):
        """Apply 1/1-octaves bandpass filters.

        :param frequencies: Band-pass filter frequencies.
        :type frequencies: :class:`np.ndarray` with (approximate) center-frequencies or an instance of :class:`acoustics.signal.Frequencies`
        :param order: Filter order.
        :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Frequencies and band-pass filtered signal.

        .. seealso:: :func:`acoustics.signal.bandpass_octaves`
        """
        frequencies, octaves = acoustics.signal.bandpass_octaves(self, self.fs, frequencies, order, purge,
                                                                 zero_phase=zero_phase)
        return frequencies, type(self)(octaves, self.fs)

    def third_octaves(self, frequencies=NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES, order=8, purge=True, zero_phase=False):
        """Apply 1/3-octaves bandpass filters.

        :param frequencies: Band-pass filter frequencies.
        :type frequencies: :class:`np.ndarray` with (approximate) center-frequencies or an instance of :class:`acoustics.signal.Frequencies`
        :param order: Filter order.
        :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Frequencies and band-pass filtered signal.

        .. seealso:: :func:`acoustics.signal.bandpass_third_octaves`
        """
        frequencies, octaves = acoustics.signal.bandpass_third_octaves(self, self.fs, frequencies, order, purge,
                                                                       zero_phase=zero_phase)
        return frequencies, type(self)(octaves, self.fs)

    def fractional_octaves(self, frequencies=None, fraction=1, order=8, purge=True, zero_phase=False):
        """Apply 1/N-octaves bandpass filters.

        :param frequencies: Band-pass filter frequencies.
        :type frequencies: Instance of :class:`acoustics.signal.Frequencies`
        :param fraction: Default band-designator of fractional-octaves.
        :param order: Filter order.
        :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Frequencies and band-pass filtered signal.

        .. seealso:: :func:`acoustics.signal.bandpass_fractional_octaves`
        """
        if frequencies is None:
            frequencies = acoustics.signal.OctaveBand(fstart=NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES[0],
                                                      fstop=self.fs / 2.0, fraction=fraction)
        frequencies, octaves = acoustics.signal.bandpass_fractional_octaves(self, self.fs, frequencies, fraction, order,
                                                                            purge, zero_phase=zero_phase)
        return frequencies, type(self)(octaves, self.fs)

    def plot_octaves(self, **kwargs):
        """Plot octaves.

        .. seealso:: :meth:`octaves`

        """
        params = {
            'xscale': 'log',
            'yscale': 'linear',
            'xlabel': '$f$ in Hz',
            'ylabel': '$L_{p}$ in dB',
            'title': '1/1-Octaves SPL',
        }
        params.update(kwargs)
        f, o = self.octaves()
        print(len(f.center), len(o.leq()))
        return _base_plot(f.center, o.leq().T, params)

    def plot_third_octaves(self, **kwargs):
        """Plot 1/3-octaves.

        .. seealso:: :meth:`third_octaves`

        """
        params = {
            'xscale': 'log',
            'yscale': 'linear',
            'xlabel': '$f$ in Hz',
            'ylabel': '$L_{p}$ in dB',
            'title': '1/3-Octaves SPL',
        }
        params.update(kwargs)
        f, o = self.third_octaves()
        return _base_plot(f.center, o.leq().T, params)

    def plot_fractional_octaves(self, frequencies=None, fraction=1, order=8, purge=True, zero_phase=False, **kwargs):
        """Plot fractional octaves.
        """
        title = '1/{}-Octaves SPL'.format(fraction)

        params = {
            'xscale': 'log',
            'yscale': 'linear',
            'xlabel': '$f$ in Hz',
            'ylabel': '$L_p$ in dB',
            'title': title,
        }
        params.update(kwargs)
        f, o = self.fractional_octaves(frequencies=frequencies, fraction=fraction, order=order, purge=purge,
                                       zero_phase=zero_phase)
        return _base_plot(f.center, o.leq().T, params)

    def plot(self, **kwargs):
        """Plot signal as function of time. By default the entire signal is plotted.

        :param filename: Name of file.
        :param start: First sample index.
        :type start: Start time in seconds from start of signal.
        :param stop: Last sample index.
        :type stop: Stop time in seconds. from stop of signal.
        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': '$t$ in s',
            'ylabel': '$x$ in -',
            'title': 'Signal',
        }
        params.update(kwargs)
        return _base_plot(self.times(), self, params)

    #def plot_scalo(self, filename=None):
    #"""
    #Plot scalogram
    #"""
    #from scipy.signal import ricker, cwt

    #wavelet = ricker
    #widths = np.logspace(-1, 3.5, 10)
    #x = cwt(self, wavelet, widths)

    #interpolation = 'nearest'

    #from matplotlib.ticker import LinearLocator, AutoLocator, MaxNLocator
    #majorLocator = LinearLocator()
    #majorLocator = MaxNLocator()

    #fig = plt.figure()
    #ax = fig.add_subplot(111)
    #ax.set_title('Scaleogram')
    ##ax.set_xticks(np.arange(0, x.shape[1])*self.fs)
    ##ax.xaxis.set_major_locator(majorLocator)

    ##ax.imshow(10.0 * np.log10(x**2.0), interpolation=interpolation, aspect='auto', origin='lower')#, extent=[0, 1, 0, len(x)])
    #ax.pcolormesh(np.arange(0.0, x.shape[1])/self.fs, widths, 10.0*np.log(x**2.0))
    #if filename:
    #fig.savefig(filename)
    #else:
    #return fig

    #def plot_scaleogram(self, filename):
    #"""
    #Plot scaleogram
    #"""
    #import pywt

    #wavelet = 'dmey'
    #level = pywt.dwt_max_level(len(self), pywt.Wavelet(wavelet))
    #print level
    #level = 20
    #order = 'freq'
    #interpolation = 'nearest'

    #wp = pywt.WaveletPacket(self, wavelet, 'sym', maxlevel=level)
    #nodes = wp.get_level(level, order=order)
    #labels = [n.path for n in nodes]
    #values = np.abs(np.array([n.data for n in nodes], 'd'))

    #fig = plt.figure()
    #ax = fig.add_subplot(111)
    #ax.set_title('Scaleogram')
    #ax.imshow(values, interpolation=interpolation, aspect='auto', origin='lower', extent=[0, 1, 0, len(values)])
    ##ax.set_yticks(np.arange(0.5, len(labels) + 0.5))
    ##ax.set_yticklabels(labels)

    #fig.savefig(filename)

    def normalize(self, gap=6.0, inplace=False):
        """Normalize signal.

        :param gap: Gap between maximum value and ceiling in decibel.
        :param inplace: Normalize signal in place.

        The parameter `gap` can be understood as using `gap` decibels fewer for the dynamic range.
        By default a 6 decibel gap is used.

        """
        factor = (self.max() * 10.0**(gap / 20.0))
        if inplace:
            self /= factor
            return self
        else:
            return self / factor

    def to_wav(self, filename, depth=16):
        """Save signal as WAV file.

        :param filename: Name of file to save to.
        :param depth: If given, convert to integer with specified depth. Else, try to store using the original data type.

        By default, this function saves a normalized 16-bit version of the signal with at least 6 dB range till clipping occurs.

        """
        data = self
        dtype = data.dtype if not depth else 'int' + str(depth)
        if depth:
            data = (data * 2**(depth - 1) - 1).astype(dtype)
        wavfile.write(filename, int(self.fs), data.T)
        #wavfile.write(filename, int(self.fs), self._data/np.abs(self._data).max() *  0.5)
        #wavfile.write(filename, int(self.fs), np.int16(self._data/(np.abs(self._data).max()) * 32767) )

    @classmethod
    def from_wav(cls, filename):
        """
        Create an instance of `Signal` from a WAV file.

        :param filename: Filename

        """
        fs, data = wavfile.read(filename)
        data = data.astype(np.float32, copy=False).T
        data /= np.max(np.abs(data))
        return cls(data, fs=fs)


_PLOTTING_PARAMS = {
    'title': None,
    'xlabel': None,
    'ylabel': None,
    'xscale': 'linear',
    'yscale': 'linear',
    'xlim': (None, None),
    'ylim': (None, None),
    'labels': None,
    'linestyles': ['-', '-.', '--', ':'],
}


def _get_plotting_params():
    d = dict()
    d.update(_PLOTTING_PARAMS)
    return d


def _base_plot(x, y, given_params):
    """Common function for creating plots.

    :returns: Axes object.
    :rtype: :class:`matplotlib.Axes`
    """

    params = _get_plotting_params()
    params.update(given_params)

    linestyles = itertools.cycle(iter(params['linestyles']))

    # Check if an axes object is passed in. Otherwise, create one.
    ax0 = params.get('ax', plt.figure().add_subplot(111))

    ax0.set_title(params['title'])
    if y.ndim > 1:
        for channel in y:
            ax0.plot(x, channel, linestyle=next(linestyles))
    else:
        ax0.plot(x, y)
    ax0.set_xlabel(params['xlabel'])
    ax0.set_ylabel(params['ylabel'])
    ax0.set_xscale(params['xscale'])
    ax0.set_yscale(params['yscale'])
    ax0.set_xlim(params['xlim'])
    ax0.set_ylim(params['ylim'])

    if params['labels'] is None and y.ndim > 1:
        params['labels'] = np.arange(y.shape[-2]) + 1
    if params['labels'] is not None:
        ax0.legend(labels=params['labels'])

    return ax0


__all__ = ["Signal"]



import pytest

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal

from acoustics.descriptors import leq, sel, lw, lden, ldn

@pytest.fixture
def levels():
    return [80, 80.3, 80.6, 80.4, 79.7, 76.9]

@pytest.fixture
def lday():
    return np.array([80, 70])

@pytest.fixture
def levening():
    return np.array([65, 55])

@pytest.fixture
def lnight():
    return np.array([45, 55])

@pytest.fixture
def power():
    return np.array([1e-3, 1e-5, 1e-7])

def test_leq(levels):
    calculated = leq(levels)
    real = 79.806989166
    assert_almost_equal(calculated, real)


def test_leq_int_time(levels):
    calculated = leq(levels, int_time=1/8)
    real = 88.837889036
    assert_almost_equal(calculated, real)


def test_sel(levels):
    calculated = sel(levels)
    real = 87.588501670
    assert_almost_equal(calculated, real)


def test_lw_int(power):
    calculated = lw(power[0])
    real = 90
    assert calculated == real


def test_lw_1darray(power):
    calculated = lw(power)
    real = np.array([90, 70, 50])
    assert_array_almost_equal(calculated, real)


def test_lw_1darray_wref(power):
    calculated = lw(power, Wref=1e-11)
    real = np.array([80, 60, 40])
    assert_array_almost_equal(calculated, real)


def test_lden_float(lday, levening, lnight):
    calculated = lden(lday[0], levening[0], lnight[0])
    real = 77.14095579
    assert_almost_equal(calculated, real)


def test_lden_array(lday, levening, lnight):
    calculated = lden(lday, levening, lnight)
    real = np.array([77.14095579, 67.93843392])
    assert_array_almost_equal(calculated, real)


def test_ldn_float(lday, lnight):
    calculated = ldn(lday[0], lnight[0])
    real = 77.96703252
    assert_almost_equal(calculated, real)


def test_ldn_array(lday, lnight):
    calculated = ldn(lday, lnight)
    real = np.array([77.96703252, 68.71330861])
    assert_array_almost_equal(calculated, real)



"""
Utils
=====

"""
import numpy as np
from acoustics.decibel import dbsum

SOUNDSPEED = 343.0
"""
Speed of sound in air.
"""

esum = dbsum


def mean_tl(tl, surfaces):
    """Mean tl."""
    try:
        tau_axis = tl.ndim - 1
    except AttributeError:
        tau_axis = 0
    tau = 1.0 / (10.0**(tl / 10.0))
    return 10.0 * np.log10(1.0 / np.average(tau, tau_axis, surfaces))


def wavelength(freq, c=SOUNDSPEED):
    """
    Wavelength for one or more frequencies (as ``NumPy array``).
    """
    return c / freq


def w(freq):
    """
    Angular frequency for one o more frequencies (as ``NumPy array``).
    """
    return 2.0 * np.pi * freq


def _is_1d(given):
    if isinstance(given, (int, float)):
        return given
    elif given.ndim == 1:
        return np.array([given])
    elif given.ndim == 2 and given.shape[0] == 1:
        return given[0]
    else:
        return given



import numpy as np
from acoustics.atmosphere import Atmosphere

import sys
sys.path.append('..')
from get_data_path import data_path


class TestAtmosphere:
    def test_standard_atmosphere(self):

        a = Atmosphere()
        """Default values."""
        assert (a.temperature == 293.15)
        assert (a.pressure == 101.325)
        assert (a.relative_humidity == 0.0)
        """Calculated values belonging to default values."""
        assert (abs(a.soundspeed - 343.2) < 1.0e-9)
        assert (abs(a.saturation_pressure - 2.33663045) < 1.0e-8)
        assert (abs(a.molar_concentration_water_vapour - 0.0) < 1.0e-9)
        assert (abs(a.relaxation_frequency_nitrogen - 9.0) < 1.0e-9)
        assert (abs(a.relaxation_frequency_oxygen - 24.0) < 1.0e-9)

    def test_attenuation_coefficient(self):
        """Test data is all at an air pressure of one standard atmosphere, 101.325 Pa."""
        data = np.loadtxt(data_path() + 'absorption_coefficient.csv', skiprows=1, delimiter=',')

        f = np.array([
            50.0, 63.0, 80.0, 100.0, 125.0, 160.0, 200.0, 250.0, 315.0, 400.0, 500.0, 630.0, 800.0, 1000.0, 1250.0,
            1600.0, 2000.0, 2500.0, 3150.0, 4000.0, 5000.0, 6300.0, 8000.0, 10000.0
        ])

        for row in data:
            temperature = 273.15 + row[0]  # Degrees Celsius to Kelvin
            relative_humidity = row[1]
            alpha = row[2:] / 1000.0  # Given in dB/km while we calculate in dB/m.

            assert (f.shape == alpha.shape)

            a = Atmosphere(temperature=temperature, relative_humidity=relative_humidity)
            calculated_alpha = a.attenuation_coefficient(f)

            np.testing.assert_array_almost_equal(alpha, calculated_alpha, decimal=2)



import numpy as np
from numpy.testing import assert_almost_equal

import pytest

from acoustics.power import lw_iso3746


@pytest.mark.parametrize("background_noise, expected", [
    (79, 91.153934187),
    (83, 90.187405234),
    (88, 88.153934187),
])
def test_lw_iso3746(background_noise, expected):
    LpAi = np.array([90, 90, 90, 90])
    LpAiB = background_noise * np.ones(4)
    S = 10
    alpha = np.array([0.1, 0.1, 0.1, 0.1, 0.1, 0.1])
    surfaces = np.array([10, 10, 10, 10, 10, 10])
    calculated = lw_iso3746(LpAi, LpAiB, S, alpha, surfaces)
    assert_almost_equal(calculated, expected)



import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal

import pytest

from acoustics.building import (rw_curve, rw, rw_c, rw_ctr, stc, stc_curve,
                                mass_law)

@pytest.fixture
def tl():
    return np.array([10, 11, 12, 13, 15, 18, 20, 20, 25, 22, 18, 15,
                            20, 25, 28, 30])


def test_rw_curve(tl):
    calculated = rw_curve(tl)
    real = np.array([3, 6, 9, 12, 15, 18, 21, 22, 23, 24, 25, 26, 26,
                     26, 26, 26])
    assert_array_almost_equal(calculated, real)


def test_rw(tl):
    calculated = rw(tl)
    real = 22
    assert_almost_equal(calculated, real)


def test_rw_c(tl):
    calculated = rw_c(tl)
    real = 19.657473180
    assert_almost_equal(calculated, real)


def test_rw_ctr(tl):
    calculated = rw_ctr(tl)
    real = 18.036282792
    assert_almost_equal(calculated, real)


def test_stc(tl):
    tl = np.array([18, 18, 17, 17, 20, 21, 22, 23, 25, 26, 26, 25, 25, 26,
                   30, 31])
    calculated = stc(tl)
    real = 25
    assert_almost_equal(calculated, real)


def test_stc_curve(tl):
    tl = np.array([18, 18, 17, 17, 20, 21, 22, 23, 25, 26, 26, 25, 25, 26,
                   30, 31])
    calculated = stc_curve(tl)
    real = np.array([9, 12, 15, 18, 21, 24, 25, 26, 27, 28, 29, 29, 29, 29,
                     29, 29])
    assert_array_almost_equal(calculated, real)


@pytest.mark.parametrize("freq, vol_density, thickness, theta, c, rho0, \
                         expected", [
    (200, 2000, 0.005, 0, 343, 1.225, 23.514371494),
    (150, 8000, 0.002, 20, 345, 1.18, 24.827231087),
    (np.array([200, 150]), 8000, 0.002, 20, 345, 1.18,
     np.array([27.319749023, 24.827231087])),
])
def test_mass_law(freq, vol_density, thickness, theta, c, rho0, expected):
    calculated = mass_law(freq, vol_density, thickness, theta, c, rho0)
    assert_array_almost_equal(calculated, expected)



import pkg_resources
# We use setup.cfg for which we need a 'recent' setuptools version
pkg_resources.require('setuptools>=30.3.0')

import setuptools
from setuptools.command.test import test as TestCommand
import sys


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


setuptools.setup(
    cmdclass={'test': PyTest},
)
